//
//  ViewController.swift
//  DIM 3.0
//
//  Created by G.J. Parker on 19/1/17.
//  Copyright © 2019 G.J. Parker. All rights reserved.
//

import Cocoa
import AppleScriptObjC

class ViewController: NSViewController {
    
    // variables to deal w/ different Arrangements and what the code should do (these are default values which will be overwritten soon)
    var restoreAtStart = false  // Restore icon positions at start?
    var quitAfterStart = false  // if we are Restoring at start should we just quit afterwards?
    var currentName = "Default" // some name for an icon Arrangment
    var arrangements = [String: Any]()  // dictionary keyed to name w/ corresponding iconSet (AppleScript data object)
    var orderedArrangements = [String]()  // an ordered list of Arrangement names to populate drop down menu and Edit sheet
    
    // these are disposable run variables
    var start = true     // did we just start?
    var overrideSetting = false  // is user holding command (⌘) during start?
    
    
    // our outlets to various labels, buttons, etc on the main storyboard
    @IBOutlet weak var doingTF: NSTextField!
    @IBOutlet weak var doingPI: NSProgressIndicator!
    @IBOutlet weak var warningTF: NSTextField!
    @IBOutlet weak var warningButton: NSButton!
    @IBOutlet weak var quitButton: NSButton!
    @IBOutlet weak var currentTF: NSTextField!
    @IBOutlet weak var currentNumDesktop: NSTextField!
    @IBOutlet weak var currentNumArrangement: NSTextField!
    @IBOutlet weak var arrangementButton: NSPopUpButton!
    @IBOutlet weak var restoreButton: NSButton!
    @IBOutlet weak var memorizeButton: NSButton!
    
    var DIM: DIMBridge? // the Bridge to AppleScript
    
    
    override func viewDidLoad() {
        super.viewDidLoad()
        overrideSetting = NSEvent.modifierFlags == .command  // check to see if user is holding command key during launch
    }
    
    override func viewDidAppear() {
        super.viewDidAppear()
        if start {  // first time through?
            if DIM == nil { //inject AppleScript stuff... if we haven't yet
                Bundle.main.loadAppleScriptObjectiveCScripts()
                let DIMBridgeClass: AnyClass = NSClassFromString("ApplescriptBridge")!
                DIM = DIMBridgeClass.alloc() as? DIMBridge
            }
            if DIM!.testBridge == nil {
                errorwithAS()
            } else {
                loadPrefs()  // load user preferences and get going
                start = false // flag to say we did this once, no need to do it again (since viewDidAppear can be called again, say, if app was Hid or not...
            }
        }
    }
    
    override var representedObject: Any? {
        didSet {
            // Update the view, if already loaded.
        }
    }
    
    
    // Button pressed to Memorize...
    @IBAction func do_memorize(_ sender: NSButton) {
        memorize(currentName)
    }
    
    // Button pressedd to Restore...
    @IBAction func do_restore(_ sender: NSButton) {
        restore(currentName)
    }
    
    // Check for Restore at start was toggled
    @IBAction func restoreQuitCheck(_ sender: NSButton) {
        restoreAtStart = (sender.state.rawValue == 1)             // "1" is checked, return true in that case
        warningTF.isHidden = !(restoreAtStart && quitAfterStart)  // if this is checked AND quit after Restore, warn the user
        quitButton.isEnabled = restoreAtStart                     // enable Quit after Restore if latter is checked
        savePrefs()
    }
    
    // Check for Quit after Restore was toggled
    @IBAction func quitThenCheck(_ sender: NSButton) {
        quitAfterStart = (sender.state.rawValue == 1)             // "1" is checked
        warningTF.isHidden = !(restoreAtStart && quitAfterStart)  // if this is checked (we know Restore is), warn the user
        savePrefs()
    }
    
    // memorize an arrangement given by name
    func memorize(_ name: String) {
        updateUI("Memorizing Icon Positions...")
        DispatchQueue.global(qos: .utility).async { [unowned self] in  // apparently we need to do this otherwise UI isn't updated during AppleScript call
            self.arrangements[name] = self.fetchSet()
            DispatchQueue.main.async {
                self.savePrefs()
                self.updateUI()
            }
        }
    }
    
    // restore from arrangement given by name
    func restore(_ name: String) {
        updateUI("Restoring Icon Positions...")
        DispatchQueue.global(qos: .utility).async { [unowned self] in
            self.setSet(set: self.arrangements[name]!)
            self.DIM!.numOnDesktop()
            DispatchQueue.main.async {
                self.updateUI()
            }
        }
    }
    
    // AppleScript can take a while, turn off all controlls until it's done and do an animation so user doesn't get too confused
    func updateUI(_ message: String) {
        doingTF.stringValue = message
        memorizeButton.isEnabled = false
        restoreButton.isEnabled = false
        doingTF.isHidden = false
        doingPI.startAnimation(nil)
        warningButton.isEnabled = false
        quitButton.isEnabled = false
        warningTF.isHidden = true
        arrangementButton.isEnabled = false
    }
    
    // apparently AppleScript call is done, turn on the contollers again so the user can do something...
    func updateUI() {  //turn on controllers once AppleScript is done
        memorizeButton.isEnabled = true
        restoreButton.isEnabled = true
        doingTF.isHidden = true
        doingPI.stopAnimation(nil)
        warningButton.isEnabled = true
        quitButton.isEnabled = restoreAtStart
        warningTF.isHidden = !(restoreAtStart && quitAfterStart)
        arrangementButton.isEnabled = true
        updateInfo()
    }
    
    //set AppleScript data and restore icon positions
    func setSet(set: Any) {
        DIM!.iconSet = set
        DIM!.restore()
    }
    
    // restore icon positions w/ AppleScript data currently being held
    func setSet() {
        DIM!.restore()    }
    
    // memorize icon positions and return AppleScript data
    func fetchSet() -> Any {
        DIM!.memorize()
        return DIM!.iconSet
    }
    
    // let's read user's preferences
    func loadPrefs() {  // go read user's perferences
        if goodLoadPrefs() {  // try to be robust in reading them back in, if there is an unrecoverable problem (or data doesn't exist), start afresh
            if restoreAtStart && !overrideSetting {  // did they want us to restore automatically at start?
//              restore(currentName) //doesn't seem to work, so just brute force a restore  (instead of the next line)
                setSet(set: arrangements[currentName]!)
                if quitAfterStart {
                    exit(0)
                } else {
                    DIM!.numOnDesktop()  // we have to make sure numArrangement, numDesktop and iconSet is set, if we got here, we only have to update numDesktop so tell Finder to do so
                }
            } else {
                DIM!.iconSet = arrangements[currentName]!  // no automatic restore, so just load AppleScript data (iconSet, numDesktop and numSet) for current arrangment
                DIM!.numInSet()     // tell Finder to compute
                DIM!.numOnDesktop() // tell Finder to compute
            }
        } else {   // first run or no user data or fatal problem reading user data
            restoreAtStart = false  // these are the defaults
            quitAfterStart = true
            currentName = "Default"
            orderedArrangements.append(currentName)
//          memorize(currentName)  // this should have worked, but it doesn't, just go do it even if the user has no feedback... (next line)
            arrangements[currentName] = fetchSet()  // this will cause numArrangment, numDesktop and iconSet to be defined so we are all good, it also stores the initial arrangement- we always need at least one!
            savePrefs()
            doingTF.isHidden = true
        }
        warningButton.state = (restoreAtStart ? .on : .off) // set default state of check for Restore at start
        quitButton.state = (quitAfterStart ? .on : .off)    // set default state of check for Quit after Restore at start
        quitButton.isEnabled = restoreAtStart               // if Restore at startup, allow the user to pick if we should quit or not, if not Restore at start, then don't allow the user to edit this
        warningTF.isHidden = !(restoreAtStart && quitAfterStart) // if Restore and Quit at startup, warn the user how to get back to our screen
        
        loadMenu()  // finally, construct the arrangement drop down menu
    }
    
    //let's assume something bad happened to the stored user data...
    func goodLoadPrefs() -> Bool {
        let defaults = UserDefaults.standard
        guard let name = defaults.string(forKey: "currentName")  else { return false }  // is there a plist?
        guard (defaults.array(forKey: "orderedArrangements") != nil) else { return false }
        currentName = name
        quitAfterStart = defaults.bool(forKey: "quitAfterStart")
        restoreAtStart = defaults.bool(forKey: "restoreAtStart")
        orderedArrangements = defaults.array(forKey: "orderedArrangements") as! [String]
        arrangements = defaults.dictionary(forKey: "arrangements")!
        if defaults.string(forKey: "donate") != nil {donateLabel.textColor = NSColor.labelColor}
        
        // in a perfect world we would be done. but let's not assume perfect and instead assume non-perfect
        // first, let's construct a new array using the data we (supposedly) have in arrangements dictionary
        var valid = [String]()
        for (arr, data) in arrangements {
            if let dictEntry = data as? NSArray {
                if dictEntry.count > 4 {valid.append(arr)} else { arrangements[arr] = nil}
            }
        }
        var corrected = [String]() // now try to preserve the order, valid has all 'valid' arrangements
        for arr in orderedArrangements {
            if valid.contains(arr) {corrected.append(arr)}
        }
        if corrected.count < valid.count {  // and assume orderedArrangement was corrupt, just add what's missing
            for arr in valid {
                if !corrected.contains(arr) {corrected.append(arr)}
            }
        }
        if corrected.count > 0 {  // we have something, if not all
            if arrangements[currentName] == nil {  // and confirm we have a valid currentName
                currentName = corrected[0]  // oh no, we don't- grab the first one and reset flag for restoring at start
                restoreAtStart = false
            }
            orderedArrangements = corrected  // either everything was perfect or we corrected what wasn't.
            return true  // and tell them to use this data
        }
        arrangements.removeAll()
        orderedArrangements.removeAll()
        return false // we don't have sets recoverable, punt and redo
    }
    
    // save user's preferences
    func savePrefs() {
        let defaults = UserDefaults.standard
        defaults.set(currentName, forKey: "currentName")
        defaults.set(restoreAtStart, forKey: "restoreAtStart")
        defaults.set(quitAfterStart, forKey: "quitAfterStart")
        defaults.set(orderedArrangements, forKey: "orderedArrangements")
        defaults.set(arrangements, forKey: "arrangements")
    }
    
    // construct the Arrangement popdown menu
    func loadMenu() {
        updateInfo()
        if let nn = arrangementButton.menu?.numberOfItems {  // destroy the existing menus
            for num in 1 ..< nn {
                arrangementButton.menu?.removeItem(at: nn-num)
            }
            for name in orderedArrangements { // create a new menu
                let menuItem = NSMenuItem(title: name, action: nil, keyEquivalent: "")
                menuItem.state = (name == currentName ? .on : .off)
                arrangementButton.menu?.addItem(menuItem)
            }
            arrangementButton.menu?.addItem(NSMenuItem.separator()) // simple seperator
            let editMenu = NSMenuItem(title: "Edit...", action: #selector(editArrangement), keyEquivalent: "") // and finally the "Edit..." option
            arrangementButton.menu?.addItem(editMenu)
            arrangementButton.menu?.addItem(NSMenuItem.separator()) // simple seperator
            let showMenu = NSMenuItem(title: "Select unmemorized Icons", action: #selector(showNewIcons), keyEquivalent: "")
            arrangementButton.menu?.addItem(showMenu)
        }
    }
    
    func updateInfo() {
        currentTF.stringValue = "Use Icon Arrangement: " + currentName  // some useful(?) info for user
        if let num = DIM!.numArrangement as? Int {currentNumArrangement.stringValue = "Number of memorized Desktop Icons: " + String(num)} //AppleScript data should be insync
        if let num = DIM!.numDesktop as? Int {currentNumDesktop.stringValue = "Number of current Desktop Icons: " + String(num)}
    }
    
    // user selected a different(?) arrangement
    @IBAction func arrangeButton(_ sender: NSPopUpButton) {
        if let name = sender.selectedItem?.title {
            if name != currentName {
                currentName = name
                DIM!.iconSet = arrangements[currentName]!  // sync AppleScript data
                DIM!.numInSet()
                savePrefs()
                loadMenu()
            }
        }
    }
    
    // user wants to highlight new icons...
    @objc func showNewIcons(_ sender: NSMenuItem) {
        DIM!.showNewIcons()
    }
    
    // user wants to edit arrangements...
    @objc func editArrangement(_ sender: NSMenuItem) {
        performSegue(withIdentifier: "toEditSheet", sender: self)
    }
    
    // give EditSheet access to the global data (could use notifications, but...)
    override func prepare(for segue: NSStoryboardSegue, sender: Any?) {
        if segue.identifier == "toEditSheet" {
            if let dvc = segue.destinationController as? EditSheet {
                dvc.myContainerViewDelegate = self
            }
        }
    }
    
    @IBOutlet weak var donateLabel: NSTextField!
    // hardcoded URL for donations (oh please!)
    @IBAction func donateClicked(_ sender: NSButton) {
        let url = URL(string: "http://www.parker9.com/d")
        NSWorkspace.shared.open(url!)
        donateLabel.textColor = NSColor.systemGray  //labelColor.withAlphaComponent(0.2)
        UserDefaults.standard.set("done", forKey: "donate")
    }
    
    // hardcoded URL for home
    @IBAction func homeClicked(_ sender: NSButton) {
        let url = URL(string: "http://www.parker9.com/desktopIconManager3.html#d")
        NSWorkspace.shared.open(url!)
    }
    
    func errorwithAS() {
        performSegue(withIdentifier: "toError", sender: self)
    }
}


